/* ARMCI header file */
#ifndef _ARMCI_H
#define _ARMCI_H   

typedef struct {
    void **src_ptr_array;
    void **dst_ptr_array;
    int  ptr_array_len;
    int bytes;
} armci_giov_t;

#define ARMCI_Put(src, dst, bytes, proc) \
        ARMCI_PutS((src), NULL, (dst), NULL, &(bytes), 0, proc)
#define ARMCI_Get(src, dst, bytes, proc) \
        ARMCI_GetS((src), NULL, (dst), NULL, &(bytes), 0, proc)

extern int ARMCI_Init(void);    /* initialize ARMCI */
extern int ARMCI_PutS(          /* strided put */
                void *src_ptr,        /* pointer to 1st segment at source*/ 
		int src_stride_arr[], /* array of strides at source */
		void* dst_ptr,        /* pointer to 1st segment at destination*/
		int dst_stride_arr[], /* array of strides at destination */
		int count[],          /* number of units at each stride level count[0]=bytes */
		int stride_levels,    /* number of stride levels */
                int proc	      /* remote process(or) ID */
                );

extern int ARMCI_AccS(                /* strided accumulate */
                int  optype,          /* operation */
                void *scale,          /* scale factor x += scale*y */
                void *src_ptr,        /* pointer to 1st segment at source*/ 
		int src_stride_arr[], /* array of strides at source */
		void* dst_ptr,        /* pointer to 1st segment at destination*/
		int dst_stride_arr[], /* array of strides at destination */
		int count[],          /* number of units at each stride level count[0]=bytes */
		int stride_levels,    /* number of stride levels */
                int proc	      /* remote process(or) ID */
                );



extern int ARMCI_GetS(          /* strided get */
                void *src_ptr,        /* pointer to 1st segment at source*/ 
		int src_stride_arr[], /* array of strides at source */
		void* dst_ptr,        /* pointer to 1st segment at destination*/
		int dst_stride_arr[], /* array of strides at destination */
		int count[],          /* number of units at each stride level count[0]=bytes */
		int stride_levels,    /* number of stride levels */
                int proc	      /* remote process(or) ID */
                );

extern int ARMCI_GetV( armci_giov_t darr[], /* descriptor array */
                int len,  /* length of descriptor array */
                int proc  /* remote process(or) ID */
              );

extern int ARMCI_PutV( armci_giov_t darr[], /* descriptor array */
                int len,  /* length of descriptor array */
                int proc  /* remote process(or) ID */
              );

extern int ARMCI_AccV( int op,       /* operation code */
                void *scale,         /* scaling factor for accumulate */
                armci_giov_t darr[], /* descriptor array */
                int len,             /* length of descriptor array */
                int proc             /* remote process(or) ID */
              );


extern int ARMCI_Malloc(void* ptr_arr[], int bytes);
extern int ARMCI_Free(void *ptr);

extern int ARMCI_MPI_Malloc(MPI_Comm comm, void *ptr_arr[],int bytes) ;
extern int ARMCI_MPI_Free(MPI_Comm comm, void *ptr) ;


extern void ARMCI_Finalize();    /* terminate ARMCI */
extern void ARMCI_Error(char *msg, int code);
extern void ARMCI_Fence(int proc);
extern void ARMCI_AllFence(void);
extern int  ARMCI_Rmw(int op, int *ploc, int *prem, int extra, int proc);
extern void ARMCI_Cleanup(void);

#define FAIL  -1
#define FAIL2 -2
#define FAIL3 -3
#define FAIL4 -4
#define FAIL5 -5
#define FAIL6 -6
#define FAIL7 -7
#define FAIL8 -8
#define ARMCI_FETCH_AND_ADD 88
#define ARMCI_FETCH_AND_ADD_LONG 89
#define ARMCI_ACC_INT 11
#define ARMCI_ACC_DBL 12
#define ARMCI_ACC_FLT 13
#define ARMCI_ACC_CPL 14
#define ARMCI_ACC_DCP 15
#define ARMCI_ACC_LNG 16
#endif
