/*Wed Jan 25 10:25:49 PST 1995*/
  
#if !defined(__STDC__) || !defined(__cplusplus) && !defined(LINUX)
#  define volatile
#endif

#define MAX_REG     128             /* max number of shmem regions per array */
#define RESERVED    2*sizeof(long)  /* used for shmem buffer management */  
#define FNAM        31              /* length of Fortran names   */
#define FLEN        80              /* length of Fortran strings */
#define BUF_SIZE    4096            /* size of shmem buffer */ 
#define ERR_STR_LEN 256             /* length of string for error reporting */

#ifdef  CRAY_T3D
#       define ALLIGN_SIZE      32
#else
#       define ALLIGN_SIZE      128
#endif

#define MAX_PTR MAX_NPROC
#define MAXDIM  4
#define MAPLEN  (MIN(GAnproc, MAX_NPROC) +MAXDIM)


typedef struct {
       Map maps [MAXDIM] ;      /* vector of Adlib dimension maps       */
       DAD dad ;                /* Adlib distributed array descriptor   */
       char **ptr;              /* arrays of pointers to remote data    */
       Integer type;            /* type of array                        */
       int  actv;               /* activity status                      */
       Integer size;            /* size of local data in bytes          */
       int elemsize;            /* sizeof(datatype)                     */
       long lock;               /* lock                                 */
       long id;			/* ID of shmem region / MA handle       */
       char name[FNAM+1];       /* array name                           */
} global_array_t;


static global_array_t GA[MAX_ARRAYS]; 
static int max_global_array = MAX_ARRAYS;
Integer *map;       /* used in get/put/acc */
Integer *proclist;
extern Integer in_handler;               /* set in interrupt handler*/


char err_string[ ERR_STR_LEN];        /* string for extended error reporting */
char *GA_name_stack[NAME_STACK_LEN];  /* stack for storing names of GA ops */ 
int  GA_stack_size=0;

/**************************** MACROS ************************************/

#define allign__(n, SIZE) \
        (((n)%SIZE) ? (n)+SIZE - (n)%SIZE: (n))

#define allign_size(n) allign__((long)(n), ALLIGN_SIZE)
#define allign_page(n) allign__((long)(n), PAGE_SIZE)

#define ga_check_handleM(g_a, string) \
{\
    if(GA_OFFSET+ (*g_a) < 0 || GA_OFFSET+(*g_a) >= max_global_array){ \
      sprintf(err_string, "%s: INVALID ARRAY HANDLE", string);         \
      ga_error(err_string, (*g_a));                                    \
    }\
    if( ! (GA[GA_OFFSET+(*g_a)].actv) ){                               \
      sprintf(err_string, "%s: ARRAY NOT ACTIVE", string);             \
      ga_error(err_string, (*g_a));                                    \
    }                                                                  \
}
      
/* this macro finds cordinates of the chunk of array owned by processor proc */

/* In general Adlib allows more complex strategies for mapping logical       */
/* process coordinates to physical processors than GA, and the reverse       */
/* mapping -- physical to logical -- isn't a function that can be computed   */
/* efficiently with the documented Adlib interface.  (Adlib programs         */
/* normally work in terms of logical coordinates, so this isn't usually a    */
/* problem.)                                                                 */

/* This implementation relies on the undocumented fact that, *by default*,   */
/* Adlib maps process coordinates to physical processors in the same way as  */
/* GA.                                                                       */

#define ga_ownsM(ga_handle, proc, lo, hi)                                      \
{                                                                              \
   DAD* _dad = &(GA[ga_handle].dad) ;                                          \
   const Procs* _p = _dad->grp().prc() ;                                       \
   int _dim  = _dad->rnk() ;                                                   \
   if(proc >= _p->size() || proc < 0) for(int _d = 0; _d < _dim; _d++){        \
         lo[_d] = (Integer)0;                                                  \
         hi[_d] = (Integer)-1;                                                 \
   }else{                                                                      \
         int _index = proc ;                                                   \
         for(int _d=0 ; _d<_dim; _d++){                                        \
             int _nblock = _p->dim(_d).size() ;                                \
             Block _i ;                                                        \
             _dad->rng(_d).block(&_i, _index % _nblock) ;                      \
             lo[_d] = _i.glb_bas + 1 ;                                         \
             hi[_d] = _i.glb_bas + _i.count ;                                  \
             _index /= _nblock ;                                               \
         }                                                                     \
   }                                                                           \
}


/*****************************************************************************/


extern "C" {
  #include "armci.h"
}


/* MA addressing */
DoubleComplex   *DCPL_MB;           /* double precision complex base address */
DoublePrecision *DBL_MB;            /* double precision base address */
Integer         *INT_MB;            /* integer base address */


/* cache numbers of GA/message-passing processes and ids */
static Integer GAme, GAnproc, GAmaster;
static Integer MPme, MPnproc;

static int GAinitialized = 0;
int ProcListPerm[MAX_NPROC];            /* permuted list of processes */
Integer local_buf_req=0;
Integer *NumRecReq = &local_buf_req;/* # received requests by data server */
                                    /* overwritten by shmem buf ptr if needed */
struct ga_stat_t GAstat = {0,0,0,0,0,0,0,0,0,0,0};
struct ga_bytes_t GAbytes ={0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.};
long   *GAstat_arr;  

    
#ifdef CRAY_T3D
#      include <fortran.h>
#endif

/* set total limit (bytes) for memory usage per processor to "unlimited" */ 
static Integer GA_total_memory = -1;
static Integer GA_memory_limited = 0;


#if defined(__STDC__) || defined(__cplusplus)
# define ARGS_(s) s
#else
# define ARGS_(s) ()
#endif

extern "C" {
  extern logical gaDirectAccess ARGS_((Integer, int  ));
  extern void ma_ga_get_ptr_ ARGS_((char **, char *));
  extern Integer ma_ga_diff_ ARGS_((char *, char *));
  extern void ma_ga_base_address_ ARGS_((Void*, Void**));
  extern void ga_sort_scat ARGS_((Integer*,Void*,Integer*,Integer*,Integer*, Integer));
  extern void ga_sort_gath_ ARGS_((Integer*, Integer*, Integer*, Integer*));
}


#undef ARGS_

#ifdef GA_TRACE
  static Integer     op_code;
#endif

#define FLUSH_CACHE 
