#ifndef _macdecls_h
#define _macdecls_h

/* 
 * $Id: macdecls.h,v 1.6 1998/05/26 20:15:20 d3h325 Exp $
 */

/* 
 * Public header file for a portable dynamic memory allocator.
 *
 * This file may be included by internal and external C files.
 */

#include "macommon.h"
#include "matypes.h"

/**
 ** constants
 **/

/* datatypes */
#define MT_CHAR		MT_C_CHAR	/* char */
#define MT_INT		MT_C_INT	/* int */
#define MT_LONGINT	MT_C_LONGINT	/* long int */
#define MT_FLOAT	MT_C_FLOAT	/* float */
#define MT_DBL		MT_C_DBL	/* double */
#define MT_LDBL		MT_C_LDBL	/* long double */
#define MT_SCPL		MT_C_SCPL	/* single precision complex */
#define MT_DCPL		MT_C_DCPL	/* double precision complex */
#define MT_LDCPL	MT_C_LDCPL	/* long double precision complex */

#define MT_C_FIRST	MT_CHAR		/* first type */
#define MT_C_LAST	MT_LDCPL	/* last type */

/**
 ** function types
 **/

extern void MA_trace(Integer value) ;

extern Boolean MA_alloc_get(Integer  datatype,
                            Integer  nelem,
                            char*    name,
                            Integer* memhandle,
                            Integer* index) ;

extern Boolean MA_allocate_heap(Integer datatype,
                                Integer nelem,
                                char *name,
                                Integer *memhandle) ;

extern Boolean MA_chop_stack(Integer memhandle) ;

extern Boolean MA_free_heap(Integer memhandle) ;

extern Boolean MA_get_index(Integer memhandle, Integer* index) ;

extern Boolean MA_get_next_memhandle(Integer *ithandle, Integer* memhandle) ;

extern Boolean MA_get_pointer(Integer memhandle, Pointer* pointer) ;

extern Pointer MA_get_mbase(Integer datatype) ;

extern Boolean MA_init(Integer datatype,
                       Integer nominal_stack,
                       Integer nominal_heap) ;

extern Boolean MA_initialized() ;

extern Boolean MA_init_memhandle_iterator(Integer* ithandle) ;

extern Integer MA_inquire_avail(Integer datatype) ;

extern Integer MA_inquire_heap(Integer datatype) ;

extern Integer MA_inquire_stack(Integer datatype) ;

extern Boolean MA_pop_stack(Integer memhandle) ;

extern void MA_print_stats(Boolean printroutines) ;

extern Boolean MA_push_get(Integer  datatype,
                           Integer  nelem,
                           char*    name,
                           Integer* memhandle,
                           Integer* index) ;

extern Boolean MA_push_stack(Integer  datatype,
                             Integer  nelem,
                             char*    name,
                             Integer* memhandle) ;

extern Boolean MA_set_auto_verify(Boolean value) ;

extern Boolean MA_set_error_print(Boolean value) ;

extern Boolean MA_set_hard_fail(Boolean value) ;

extern Integer MA_sizeof(Integer datatype1,
                         Integer nelem1,
                         Integer datatype2) ;

extern Integer MA_sizeof_overhead(Integer datatype) ;

extern void MA_summarize_allocated_blocks() ;

extern Boolean MA_verify_allocator_stuff() ;

/**
 ** variables
 **/

/* base arrays for the C datatypes */
extern char			ma_cb_char[];	/* MT_C_CHAR */
extern int			ma_cb_int[];	/* MT_C_INT */
extern long			ma_cb_long[];	/* MT_C_LONGINT */
extern float			ma_cb_float[];	/* MT_C_FLOAT */
extern double			ma_cb_dbl[];	/* MT_C_DBL */
extern MA_LongDouble		ma_cb_ldbl[];	/* MT_C_LDBL */
extern MA_SingleComplex		ma_cb_scpl[];	/* MT_C_SCPL */
extern MA_DoubleComplex		ma_cb_dcpl[];	/* MT_C_DCPL */
extern MA_LongDoubleComplex	ma_cb_ldcpl[];	/* MT_C_LDCPL */

#endif /* _macdecls_h */
